/* 845patch - a hack to overcome a limitation in some i845G-based laptops */
/*          - based on 855patch */

/* usage: 845patch size */
/*        size is the the new VideoRAM size in kB */

/* This source code is in the public domain. Please mention me, if you use */
/* this code in your work. */
/* This version should be quite safe to use, but don't blame me  */
/* if something bad happens. */

/* Christian Zietz <czietz@gmx.net> */

#if defined(__NetBSD__) || defined(__FreeBSD__)
#define BSD 1
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#ifndef BSD
/* in the original source the outx() macros had data, port as parameters,
   while anyone else uses port, param. Fixed it for BSD */
#define l_outb(x,y) outb(y,x)
#define l_outl(x,y) outl(y,x)

#include <sys/io.h>
#else
/* for BSD leave it as it is */
#define l_outb(x,y) outb(x,y)
#define l_outl(x,y) outl(x,y)

#if defined(__NetBSD__)
#include <i386/pio.h>
#include <i386/sysarch.h>
#define iopl(x) i386_iopl(x)
#endif
#if defined(__FreeBSD__)
#include <sys/types.h>
#include <machine/cpufunc.h>
int io=0;
int iopl(int x)
{
	printf("iopl(%d)\n",x);
	if (x>0)
	{	
		if(io==0)
		{
			if((io=open("/dev/io",O_RDWR,0))>0)
			{
				puts("opened /dev/io");
				return 0;
			}
			puts("Couldn't open /dev/io");
			io=0;
		} else
		{
			puts("/dev/io was already open");
			return 0;
		}
	} 
	if(x==0)
	{
		if(io!=0)
		{
			puts("closing /dev/io");
			close(io);
			io=0;
		}
		return 0;
	}
	return -1;
}
		
#endif
const void *memmem(const void *haystack, size_t haystacklen,
              const void *needle, size_t needlelen)
{
	int ptr=0;
	int to_go=haystacklen-needlelen;
	const unsigned char *hs=haystack;
	if(to_go>=0) // no need to do anything if needle > haystack
	{
		while(ptr<=to_go)
		{
			if(memcmp(&hs[ptr],needle,needlelen)==0)
			{
				return &hs[ptr];
			}
			ptr++;
		}
	}
	return NULL;
}
#endif
#include <unistd.h>
#define __USE_GNU
#include <sys/mman.h>

// these files come from lrmi
#include <lrmi.h>
#include <vbe.h>

#define vbiosaddr 0xc0000
#define vbiossize 0x10000
#define chipsetid 0x25608086
#define idstring "Total time for VGA POST:"
#define idoffset (-19)

int main (int argc, char *argv[]) {
  int newmemsize;

  unsigned char oldpam1, oldpam2;
  char *biosmemptr;
  int memfd = 0;

  struct LRMI_regs r;
  struct vbe_info_block *info;

  if (argc!=2) {
    fprintf(stderr, "Usage: %s size\nsize is new VideoRAM size in kB\n", argv[0]);
    return 1;
  }
  
  newmemsize = atoi(argv[1]);
  if (newmemsize==0) {
    fprintf(stderr, "Usage: %s size\nsize is new VideoRAM size in kB\n", argv[0]);
    return 1;
  }
  
  // give me permission to access io ports
  if (iopl(3)<0) {
    fprintf(stderr, "Could not set IO permissions!\n%s must be run as root\n", argv[0]);
    return 1;
  }

  // check for correct chipset
  l_outl(0xcf8, 0x80000000);
  if (inl(0xcfc) != chipsetid) {
    fprintf(stderr, "No i845G chipset found! Found chipset 0x%x instead.\n",inl(0xcfc));
  }
  printf("i845G chipset found\n");

  // access the Video BIOS
  memfd = open("/dev/mem", O_RDWR);
  if (memfd<0) {
    fprintf(stderr, "Could not open /dev/mem !\n");
    return 1;
  }

  biosmemptr = mmap((void *)vbiosaddr, vbiossize, PROT_READ | PROT_WRITE, MAP_SHARED | MAP_FIXED, memfd, vbiosaddr);
  if (biosmemptr == NULL) {
    fprintf(stderr, "Could not mmap Video BIOS!\n");
    close(memfd);
    return 1;
  }

  // search for ID string and calculate address to patch
  biosmemptr = memmem(biosmemptr, vbiossize, idstring, strlen(idstring));
  if(biosmemptr == NULL) {
    fprintf(stderr, "Unsupported Video BIOS!\n");
    munmap(biosmemptr, vbiossize);
    close(memfd);
    return 1;
  }

  biosmemptr += idoffset;

  // call BIOS and check returned memory size with that stored at the location to patch
  LRMI_init();
  info = LRMI_alloc_real(sizeof(struct vbe_info_block));
  if (info == NULL) {
    fprintf(stderr, "Could not allocate real mode memory!\n");
    munmap(biosmemptr, vbiossize);
    close(memfd);
    return 1;
  } 
   
  memset(&r, 0, sizeof(r));
  r.eax = 0x4f00;
  r.es = (unsigned int)info >> 4;
  r.edi = (unsigned int)info & 0xf;
  LRMI_int(0x10, &r);
  if (r.eax != 0x004f) {
    fprintf(stderr, "VBE call failed!\n");
    LRMI_free_real(info);
    munmap(biosmemptr, vbiossize);
    close(memfd);
    return 1;
  }

  // verify value at location to patch (rounded down) and value returned by BIOS match
  if ((*(long *)biosmemptr - 0x21000) >> 16 != info->total_memory) {
    fprintf(stderr, "Could not calculate correct location to patch!\n");
    LRMI_free_real(info);
    munmap(biosmemptr, vbiossize);
    close(memfd);
    return 1;
  }
  LRMI_free_real(info);

  // set permissions to read and write so I can write into shadowed VideoBIOS
  l_outl(0xcf8, 0x80000090);
  oldpam1 = inb(0xcfd);
  oldpam2 = inb(0xcfe);
  l_outl(0xcf8, 0x80000090);
  l_outb(0xcfd, 0x33);
  l_outb(0xcfe, 0x33);

  // patch it
  printf("BIOS currently knows of %ld kB of VideoRAM\n", *(long *)biosmemptr/1024);  
  *(long *)biosmemptr = newmemsize*1024;
  printf("%s setting size to %d kB\n", *(long *)biosmemptr/1024==newmemsize?"Successfully":"Unsuccessfully", newmemsize);  

  // reset permissions
  l_outl(0xcf8, 0x80000090);
  l_outb(0xcfd, oldpam1);
  l_outb(0xcfe, oldpam2);  

  munmap(biosmemptr, vbiossize);
  close(memfd);

  return 0;
}
